﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace TaskTest
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        #region 생성자
        public MainWindow()
        {
            InitializeComponent();

            Initialization();
        }
        #endregion

        #region 초기화 작업
        /// <summary>
        /// 초기화 작업
        /// </summary>
        private void Initialization()
        {
            Task.Factory.StartNew(() => {
                while (true)
                {
                    //Min thread work
                    int minWorker, minIOC;
                    ThreadPool.GetMinThreads(out minWorker, out minIOC);
                    tbMinThreadCount.Dispatcher.BeginInvoke(new Action(() => { tbMinThreadCount.Text = minWorker.ToString(); }), null);

                    //Max thread work
                    int maxWorker, maxIOC;
                    ThreadPool.GetMaxThreads(out maxWorker, out maxIOC);
                    tbMaxThreadCount.Dispatcher.BeginInvoke(new Action(() => { tbMaxThreadCount.Text = maxWorker.ToString(); }), null);

                    Thread.Sleep(500);
                }
            });
                
            ThreadPoolInitialization();
        }

        /// <summary>
        /// ThreadPool 관련 초기화 작업
        /// </summary>
        private void ThreadPoolInitialization()
        {
            if (rdoCustomerThreadControl.IsChecked.Value)
            {
                int minWorker, minIOC;

                minWorker = int.Parse(txtMinWorkerCount.Text);
                minIOC = int.Parse(txtMinWorkerIOCCount.Text);

                // 스레드 작성 및 소멸을 관리하기 위한 알고리즘으로 전환하기 전에 새 요청에 따라 스레드 풀이 생성하는 스레드의 최소 수를 설정합니다.
                // minWorker : 스레드 풀에서 필요할 때 만드는 작업자 I/O 스레드의 최소 개수입니다. 
                // minIOC : 스레드 풀에서 필요할 때 만드는 비동기 I/O 스레드의 최소 개수입니다. 
                ThreadPool.SetMinThreads(minWorker, minIOC);

                int maxWorker, maxIOC;
                maxWorker = int.Parse(txtMaxWorkerCount.Text);
                maxIOC = int.Parse(txtMaxWorkerIOCCount.Text);
                // 동시에 활성 상태가 될 수 있는 스레드 풀에 대한 요청 수를 설정합니다. 해당 수를 넘는 모든 요청은 스레드 풀 스레드가 사용 가능해질 때까지 큐에 대기 상태로 남아 있습니다.
                // maxWorker : 스레드 풀에 있는 최대 작업자 스레드 수입니다. 
                // maxIOC : 스레드 풀에 있는 최대 비동기 I/O 스레드 수입니다. 
                ThreadPool.SetMaxThreads(maxWorker, maxIOC);
            }
        }
        #endregion

        #region Fields
        /// <summary>
        /// 시작 등수 관리 관련 잠금 객체
        /// </summary>
        private static object startLevelLock = new object();

        /// <summary>
        /// 시작 등수 관리
        /// </summary>
        private static int startLevel = 1;

        /// <summary>
        /// 완료 등수 관리 관련 잠금 객체
        /// </summary>
        private static object completedLevelLock = new object();

        /// <summary>
        /// 완료 등수 관리
        /// </summary>
        private static int completedLevel = 1;

        /// <summary>
        /// Task를 관리할 List 객체
        /// </summary>
        /// <remarks>
        /// Task를 생성에서 바로 실행하지 않고 별도의 이벤트에서 활성화 하도록 하기 위해 만들었다.
        /// </remarks>
        private List<Task> tasks = new List<Task>();
        #endregion

        #region Methods
        /// <summary>
        /// TaskCreationOptions 옵션으로 변환해서 가져오기
        /// </summary>
        /// <param name="taskCreationOptionString"></param>
        /// <returns></returns>
        private TaskCreationOptions GetSelectedTaskCreationOption(string taskCreationOptionString)
        {
            // http://msdn.microsoft.com/ko-kr/library/system.threading.tasks.taskcreationoptions.aspx 참조
            TaskCreationOptions option = TaskCreationOptions.None;
            switch (taskCreationOptionString)
            {
                case "None":
                    // 기본 동작이 사용되도록 지정합니다.
                    option = TaskCreationOptions.None;
                    break;

                case "PreferFairness":
                    // 가능한 한 공정한 방식, 즉 일찍 예약된 작업은 일찍 실행되고 나중에 예약된 작업은 나중에 실행될 수 있는 방식으로 작업을 예약하는 TaskScheduler에 대한 힌트입니다.
                    option = TaskCreationOptions.PreferFairness;
                    break;

                case "LongRunning":
                    // 작업이 장기 실행되는 정교하지 않은 작업이 되도록 지정합니다.초과 구독을 보장할 수 있는 TaskScheduler에 대한 힌트를 제공합니다.
                    option = TaskCreationOptions.LongRunning;
                    break;

                case "AttachedToParent":
                    // 작업이 작업 계층 구조의 부모에 연결되도록 지정합니다.
                    option = TaskCreationOptions.AttachedToParent;
                    break;

                case "DenyChildAttach":
                    // 지정 하는 InvalidOperationException 만들어진된 작업에는 자식 작업을 첨부 하려고 시도 하는 경우에 throw 됩니다.
                    option = TaskCreationOptions.DenyChildAttach;
                    break;

                case "HideScheduler":
                    // 앰비언트 스케줄러를 현재 스케줄러에서 만들어진된 작업으로 표시 되지 않습니다.즉, StartNew 또는 Continuewith와 같은 생성된 작업을 수행 하는 작업을 볼 수 있도록 Default 현재 스케줄러로 합니다.
                    option = TaskCreationOptions.HideScheduler;
                    break;
            }

            return option;
        }

        /// <summary>
        /// 비 동기 수행 타스크의 취소 여부를 컨트롤 한다.
        /// </summary>
        /// <seealso cref="http://msdn.microsoft.com/ko-kr/library/dd997289.aspx"/>
        /// <seealso cref="http://msdn.microsoft.com/ko-kr/library/dd997364.aspx"/>
        private CancellationTokenSource cancellationTokenSource;

        /// <summary>
        /// 폼에 등록할 Progress을 포함한 StackPanel을 가져온다.
        /// </summary>
        /// <returns></returns>
        /// <remarks>
        /// ProgressBar의 값을 변화 시킬 수 있는 비 동기 실행을 연결 시켜 놓았다.
        /// 폼에서 Processing를 클릭하면 전체 ProgressBar의 진행바가 변경이 될 것이다.
        /// </remarks>
        private StackPanel GetProgressBar()
        {
            Parallel.Invoke(() => { }, () => { });
            // 취소 토큰 소스 초기화
            cancellationTokenSource = new CancellationTokenSource();

            // 타스트 생성 옵션 초기화
            // 콤보박스에서 선택된 값을 기준으로 변환해서 가져온다.
            TaskCreationOptions option = GetSelectedTaskCreationOption((cboTaskCreationOption.SelectedValue as ComboBoxItem).Content.ToString());

            // 천체 컨트롤을 담을 Layout 컨트롤 초기화
            var stackPanel = new StackPanel();
            // 컨트롤들이 수평으로 정렬 되도록 옵션 지정
            stackPanel.Orientation = Orientation.Horizontal;

            //ProgressBas를 세팅한다.
            var progressBar = new ProgressBar();
            progressBar.Width = 400;
            progressBar.Height = 5;
            progressBar.Minimum = 0;
            progressBar.Maximum = 10000;
            progressBar.MaxHeight = 100;
            progressBar.Value = 0;
            stackPanel.Children.Add(progressBar);

            // 시작 등수를 나타내는 컨트롤
            var textBlockStartRank = new TextBlock() { Text = "0", Margin = new Thickness(5, 0, 0, 0) };
            stackPanel.Children.Add(textBlockStartRank);

            // 완료 등수를 나타내는 컨트롤
            var textBlockCompletedRank = new TextBlock() { Text = "대기중", Margin = new Thickness(5, 0, 0, 0) };
            stackPanel.Children.Add(textBlockCompletedRank);

            Task task = new Task(() => {
                try
                {
                    // 수행 시간을 체크 하기 위한 객체
                    Stopwatch sw = new Stopwatch();
                    sw.Start();

                    int currentStartLevelLock = 0;
                    // 비 동기 수행이므로 잠김 객체로 보호해서 값을 증가 시킴
                    lock (startLevelLock)
                    {
                        // 시작 등수를 증가해서 할당 함
                        // 잠김 시간을 최대한 줄이기 위해 지역 변수를 사용 함
                        currentStartLevelLock = startLevel++;
                    }

                    // 시작 등수를 BeginInvoke를 통해 변경 함.
                    textBlockStartRank.Dispatcher.BeginInvoke(new Action(() => { textBlockStartRank.Text = currentStartLevelLock.ToString(); }));

                    // Task 수행 활성화 여부를 확인 할 수 있도록 변경
                    textBlockCompletedRank.Dispatcher.BeginInvoke(new Action(() => { textBlockCompletedRank.Text = "실행중"; }), null);

                    for (double i = 1; i <= 10000; i++)
                    {
                        // 이 토큰의 취소가 요청되었으면 true이고, 그렇지 않으면 false입니다.
                        if (cancellationTokenSource.Token.IsCancellationRequested)
                        {
                            break;
                        }

                        Thread.Sleep(1);
                        progressBar.Dispatcher.BeginInvoke(new Action(() => { progressBar.Value = i += 1; }));  //하나의 타스크가 84751ms 걸림, LongRunning 타입, 200개를 한번에 실행

                        // Invoke를 사용하지 않은 이유는 대기 시간을 줄이기 위함
                        // Invoke는 완료가 될때까지 대기해야 하기 때문에 약간의 시간이 더 걸린다.
                        // 그렇지만 적은 타스크를 동시 실행하는 환경에서는 Invoke로 실행하는 것이 더 빨리 완료되기도 한다.
                        //progressBar.Dispatcher.Invoke(new Action(() => { progressBar.Value = i += 1; }));   //하나의 타스크가 95510ms 걸림
                    }

                    int currentCompletedLevelLock = 0;
                    // 비 동기 수행이므로 잠김 객체로 보호해서 값을 증가 시킴
                    lock (completedLevelLock)
                    {
                        // 완료 등수를 증가해서 할당 함
                        // 잠김 시간을 최대한 줄이기 위해 지역 변수를 사용 함
                        currentCompletedLevelLock = completedLevel++;
                    }

                    sw.Stop();

                    textBlockCompletedRank.Dispatcher.BeginInvoke(new Action(() => { textBlockCompletedRank.Text = (currentCompletedLevelLock).ToString() + " 등 " + sw.ElapsedMilliseconds + "ms"; }));
                }
                catch (Exception ex)
                {
                    var message = ex.Message;
                }
            }, cancellationTokenSource.Token, option);

            tasks.Add(task);

            return stackPanel;
        }
        #endregion

        #region Event handler
        /// <summary>
        /// 테스트할 Task를 세팅 하는 이벤트 핸들러
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnSet_Click(object sender, RoutedEventArgs e)
        {
            // 쓰레드풀 초기화
            ThreadPoolInitialization();

            // 슬라이드 컨트롤의 값을 int로 변환
            var count = (int)currentCountSlider.Value;

            Task.Factory.StartNew(() =>
            {
                for (int i = 1; i <= count; i++)
                {
                    this.Dispatcher.BeginInvoke(new Action(() =>
                    {
                        var stackPanel = GetProgressBar();
                        // 제공된 컨트롤을 리스트 객채에 넣는다.
                        lstTaskResult.Items.Add(stackPanel);
                    }));
                }
            });
        }

        /// <summary>
        /// 모든 작업을 초기화 시키는 이벤트 핸들러
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnClear_Click(object sender, RoutedEventArgs e)
        {
            // ListItem 초기화
            lstTaskResult.Items.Clear();
            // Task를 초기화
            tasks.Clear();
            // 시작등수 초기화
            startLevel = 1;
            // 완료 등수 초기화
            completedLevel = 1;

            // ThreadPool을 초기화
            ThreadPoolInitialization();
        }

        /// <summary>
        /// 비 동기 작업을 활성화 시키는 이벤트 핸들러
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnProcessing_Click(object sender, RoutedEventArgs e)
        {
            // Processing 버튼을 눌렀을 때 비 동기 Task가 활성화 되도록 처리 함.
            Task.Factory.StartNew(() => {

                // Task를 활성화 시킴
                tasks.ForEach(task => task.Start());

                // 모두 완료가 될때까지 대기
                Task.WaitAll(tasks.ToArray());

                // 모든 Task가 수행을 마쳐 더 이상 관리가 필요 없어져서 관리에서 제거
                tasks.Clear();
            });
        }

        private void btnCancel_Click(object sender, RoutedEventArgs e)
        {
            if (cancellationTokenSource != null)
            {
                // 실행중인 타스크를 취소
                cancellationTokenSource.Cancel();
            }
        }
        #endregion
    }
}
